# S2MLCore version 1.1.1
# S2MLStochasticExpression.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Stochastic expressions

# 1) Imported modules
# -------------------

import string
import sys
import re
import math
import random

from S2MLType import *
from S2MLMetaModel import *
from S2MLDomains import *
from S2MLDistribution import *

# 2) Stochastic expressions
# -------------------------

class S2MLStochasticExpression(S2MLTerm):
  __DISTRIBUTION__ = "__DISTRIBUTION__"
  __DOMAIN__ = "__DOMAIN__"

  def __init__(self, type):
    S2MLTerm.__init__(self, type)
    self._value = S2MLConstant.NewNone()
    self._randomDeviateAbscissa = 0.5

  @staticmethod
  def NewReference(path):
    expression = S2MLStochasticExpression(S2MLType.REFERENCE_EXPRESSION)
    expression.SetReferencePath(path)
    return expression

  @staticmethod
  def NewConstant(value):
    expression = S2MLStochasticExpression(S2MLType.CONSTANT_EXPRESSION)
    expression.SetValue(value)
    return expression

  @staticmethod
  def NewWithin(child, domain):
    expression = S2MLStochasticExpression(S2MLType.WITHIN_EXPRESSION)
    expression.AppendChild(child)
    expression.SetDomain(domain)
    return expression

  @staticmethod
  def NewOperatorCall(functor, children):
    expression = S2MLStochasticExpression(S2MLType.OPERATOR_CALL_EXPRESSION)
    expression.SetFunctor(functor)
    expression.AppendChildren(children)
    return expression

  @staticmethod
  def NewOperation(type, children):
    expression = S2MLStochasticExpression(type)
    expression.AppendChildren(children)
    return expression

  def Copy(self):
    if self.GetType()==S2MLType.REFERENCE_EXPRESSION:
      myPath = self.GetReferencePath()
      path = S2MLPath.Copy(myPath)
      return S2MLStochasticExpression.NewReference(path)
    if self.GetType()==S2MLType.CONSTANT_EXPRESSION:
      myValue = self.GetValue()
      value = S2MLConstant.Copy(myValue)
      return S2MLStochasticExpression.NewConstant(value)
    if self.GetType()==S2MLType.WITHIN_EXPRESSION:
      myChild = self.GetChild(0)
      child = myChild.Copy()
      myDomain = self.GetDomain()
      domain = myDomain.Clone()
      return S2MLStochasticExpression.NewIn(child, domain)
    if self.IsEmpiricalDistribution():
      myDistribution = self.GetDistribution()
      distribution = myDistribution.Copy()
      return S2MLStochasticExpression.NewDistribution(self.GetType(), distribution)
    children = []
    for myChild in self.GetChildren():
      child = myChild.Copy()
      children.append(child)
    return S2MLStochasticExpression.NewOperation(self.GetType(), children)

  def IsValueCalculated(self):
    return not self._value.IsNone()
  def ResetValue(self):
    self._value.Reset()
  def GetValue(self):
    return self._value
  def SetValue(self, value):
    self._value.Set(value)

  def GetRandomDeviateAbscissa(self):
    return self._randomDeviateAbscissa

  def SetRandomDeviateAbscissa(self, value):
    self._randomDeviateAbscissa = value

  def IsEmpiricalDistribution(self):
    return self.GetType()==S2MLType.PIECEWISE_UNIFORM_DEVIATE_EXPRESSION \
      or self.GetType()==S2MLType.STEPWISE_DEVIATE_EXPRESSION \
      or self.GetType()==S2MLType.PIECEWISE_UNIFORM_DISTRIBUTION_EXPRESSION \
      or self.GetType()==S2MLType.STEPWISE_DISTRIBUTION_EXPRESSION
  def GetDistribution(self):
    if not self.IsEmpiricalDistribution():
      return None
    return self.GetAttribute(S2MLStochasticExpression.__DISTRIBUTION__)
  def SetDistribution(self, distribution):
    if not self.IsEmpiricalDistribution():
      return
    self.SetAttribute(S2MLStochasticExpression.__DISTRIBUTION__, distribution)

  def GetDomain(self):
    if self.GetType()!=S2MLType.WITHIN_EXPRESSION:
      return None
    return self.GetAttribute(S2MLStochasticExpression.__DOMAIN__)
  def SetDomain(self, domain):
    if self.GetType()!=S2MLType.WITHIN_EXPRESSION:
      return
    self.SetAttribute(S2MLStochasticExpression.__DOMAIN__, domain)

  def GetPriority(self):
    if self.GetType()==S2MLType.AND_EXPRESSION:
      return 8
    if self.GetType()==S2MLType.OR_EXPRESSION:
      return 9
    if self.GetType()==S2MLType.NOT_EXPRESSION:
      return 7
    if self.GetType()==S2MLType.EQ_EXPRESSION:
      return 6
    if self.GetType()==S2MLType.DF_EXPRESSION:
      return 6
    if self.GetType()==S2MLType.LT_EXPRESSION:
      return 6
    if self.GetType()==S2MLType.GT_EXPRESSION:
      return 6
    if self.GetType()==S2MLType.LEQ_EXPRESSION:
      return 6
    if self.GetType()==S2MLType.GEQ_EXPRESSION:
      return 6
    if self.GetType()==S2MLType.NEG_EXPRESSION:
      return 1
    if self.GetType()==S2MLType.ADD_EXPRESSION:
      return 5
    if self.GetType()==S2MLType.SUB_EXPRESSION:
      return 4
    if self.GetType()==S2MLType.MUL_EXPRESSION:
      return 3
    if self.GetType()==S2MLType.DIV_EXPRESSION:
      return 2
    return 0
