# Janos version 1.2.1
# JanosEngine.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of Janos, a pedagogical stochastic simulator.
# Janos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Janos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Janos.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Settings
# 4) Algorithms

# 1) Imported modules
# -------------------

from S2MLCore import *

from JanosModels import *
from JanosInterface import *
from JanosModelManager import *
from JanosInterpreter import *
from JanosSimulator import *

# 2) Creator
# ----------

class JanosEngine(S2MLEngine):

  def __init__(self):
    S2MLEngine.__init__(self)
    self._simulator = None

# 3) Settings
# -----------

  def CreateTools(self):
    self._modelManager = JanosModelManager(self)
    self._interpreter = JanosInterpreter(self)
    self._environmentManager = S2MLEnvironmentManager(self)
    self._modelManager.CreateTools()
    self._interpreter.CreateTools()
    self._environmentManager.CreateTools()
    self._environmentManager.Initialize()
    self._expressionManager = S2MLStochasticExpressionManager(self)
    self._simulator = JanosSimulator(self)

  def GetSimulator(self):
    return self._simulator

# 4) Methods to import, instantiate and export models
# ---------------------------------------------------

  def NewModel(self):
    return JanosModel()

# 3) Scripts
# ----------

  def _ParseCommand(self):
    commandName = self.LookUpArgument()
    if commandName=="compute":
      commandName = self.GetArgument()
      self._errors += self._ParseCommandCompute()
    else:
      S2MLInterpreter._ParseCommand(self)

  def ExecuteCommandCompute(self, command):
    what = command.GetArgumentAsIdentifier(0)
    if what=="observers":
      return self._ExecuteCommandComputeValuesOfObservers(command)
    return 1

  def _ExecuteCommandComputeValuesOfObservers(self, command):
    self._simulator.SetOutputFileName(None)
    targetModel = self._modelManager.GetTargetModel()
    if targetModel==None:
      self.RaiseError("No instantiated model")
      return 1
    blockName = command.GetArgumentAsIdentifier(1)
    if blockName==None:
      return 1
    targetBlock = targetModel.LookForBlock(blockName)
    if targetBlock==None:
      self.RaiseError("No block %s in instantiated model" % blockName)
      return 1

    self._simulator.SetSeed(self._environmentManager.GetValueOfEnvironmentVariable("seed").GetValueOfInteger())
    self._simulator.SetNumberOfTries(self._environmentManager.GetValueOfEnvironmentVariable("number-of-tries").GetValueOfInteger())
    # self._simulator.SetOutputFileName(self._environmentManager.GetValueOfEnvironmentVariable("output").GetValueOfString())
    # self._simulator.SetOutputFileOpeningMode(self._environmentManager.GetValueOfEnvironmentVariable("mode").GetValueOfSymbol())

    intValue = self.GetValueOfOptionOfCommandAsInteger(command, "seed", 12345)
    self._simulator.SetSeed(intValue)
    stringValue = self.GetValueOfOptionOfCommandAsString(command, "output", "results.txt")
    self._simulator.SetOutputFileName(stringValue)
    stringValue = self.GetValueOfOptionOfCommandAsFileOpeningMode(command, "mode")
    self._simulator.SetOutputFileOpeningMode(stringValue)

    missionTimes = self.GetValueOfOptionOfCommandAsListOfReals(command, "mission-times", [0.0])
    profileName = self.GetValueOfOptionOfCommandAsIdentifier(command, "profile", None)
    if profileName==None:
      profile = self._environmentManager.GetDefaultIndicatorProfile()
    else:
      profile = self._environmentManager.GetIndicatorProfile(profileName)
      if profile==None:
        self.RaiseError("line %d\nNo indicator profile %s\n" % (self._lineNumber, profileName))
        return 1
    targetBlock.SetIndicatorProfileOfObservers(profile)

    if self._simulator.GetNumberOfTries()==0:
      self._simulator.CalculateValuesOfElements(targetBlock, missionTimes)
    else:
      self._simulator.PerformStochasticSimulation(targetBlock, missionTimes)
    return 0
