# Janos version 1.2.1
# JanosModelFlattener.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of Janos, a pedagogical stochastic simulator.
# Janos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Janos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Janos.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Flattener
#   3.1) Creator
#   3.2) Instantiation of elements
#   3.3) Flattening of elements

# 1) Imported modules
# -------------------

from S2MLCore import *

from JanosModels import *
from JanosModelPrinter import *

# 3) Flattener
# ------------

# 3.1) Creator
# ------------

class JanosModelFlattener(S2MLModelFlattener):
  def __init__(self, engine):
    S2MLModelFlattener.__init__(self, engine)

# 3.2) Instantiation of elements
# -----------------------------

  def InstantiateElement(self, action):
    sourceElement = action.GetDatum(0)
    if sourceElement.GetType()==JanosType.VARIABLE:
      status = self._InstantiateVariable(action)
    elif sourceElement.GetType()==JanosType.ASSIGNMENT:
      status = self._InstantiateAssignment(action)
    elif sourceElement.GetType()==JanosType.DOUBLE_ASSIGNMENT:
      status = self._InstantiateDoubleAssignment(action)
    elif sourceElement.GetType()==JanosType.OBSERVER:
      status = self._InstantiateObserver(action)
    else:
      status = S2MLModelFlattener.InstantiateElement(self, action)
    return status

  def _InstantiateVariable(self, action):
    sourceVariable = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceVariable.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        sourceDomain = sourceVariable.GetDomain()
        targetDomain = self._CloneDomain(sourceDomain)
        targetVariable = self._targetModel.NewVariable(targetDomain, targetPath, targetContainer)
      else:
        absoluteSourcePath = sourceVariable.GetAbsolutePath()
        self.RaiseError("Variable {0:s} declared outside of its parent block\n".format(absoluteSourcePath.ToString()))
        return S2MLFlattenerAction.FAILED
    elif targetElement.GetType()==JanosType.VARIABLE:
      targetVariable = targetElement # self._targetModel.NewVariable(targetDomainIdentifier, targetPath, targetContainer)
      sourceDomain = sourceVariable.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetVariable.SetDomain(targetDomain)
    else:
      absoluteSourcePath = sourceVariable.GetAbsolutePath()
      self.RaiseError("Type mismatch for variable {0:s}\n".format(absoluteSourcePath.ToString()))
      return S2MLFlattenerAction.FAILED
    actualTargetContainer = targetVariable.GetContainer()
    for name in sourceVariable.GetAttributeNames():
      sourceExpression = sourceVariable.GetAttribute(name)
      targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, targetContainer, actualTargetContainer)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetVariable.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateAssignment(self, action):
    sourceAssignment = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceLeftMember = sourceAssignment.GetLeftMember()
    targetLeftMember = self._ContextualizedCopyOfStochasticExpression(sourceLeftMember, targetContainer, targetContainer)
    if targetLeftMember==None:
      return S2MLFlattenerAction.FAILED
    sourceRightMember = sourceAssignment.GetRightMember()
    targetRightMember = self._ContextualizedCopyOfStochasticExpression(sourceRightMember, targetContainer, targetContainer)
    if targetRightMember==None:
      return S2MLFlattenerAction.FAILED
    targetAssignment = self._targetModel.NewAssignment(targetLeftMember, targetRightMember, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateDoubleAssignment(self, action):
    sourceAssignment = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceLeftMember = sourceAssignment.GetLeftMember()
    targetLeftMember = self._ContextualizedCopyOfStochasticExpression(sourceLeftMember, targetContainer, targetContainer)
    if targetLeftMember==None:
      return S2MLFlattenerAction.FAILED
    sourceRightMember = sourceAssignment.GetRightMember()
    targetRightMember = self._ContextualizedCopyOfStochasticExpression(sourceRightMember, targetContainer, targetContainer)
    if targetRightMember==None:
      return S2MLFlattenerAction.FAILED
    targetAssignment = self._targetModel.NewDoubleAssignment(targetLeftMember, targetRightMember, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateObserver(self, action):
    sourceObserver = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceObserver.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        sourceDomain = sourceObserver.GetDomain()
        targetDomain = self._CloneDomain(sourceDomain)
        targetObserver = self._targetModel.NewObserver(targetDomain, targetPath, targetContainer)
        targetObserver.SetIdentifierAndInsertInContainerDictionary(targetPath.ToString())
      else:
        absoluteSourcePath = sourceObserver.GetAbsolutePath()
        self.RaiseError("Observer {0:s} declared outside of its parent block\n".format(absoluteSourcePath.ToString()))
        return S2MLFlattenerAction.FAILED
    elif targetElement.GetType()==S2MLType.OBSERVER:
      targetObserver = targetElement # self._targetModel.NewObserver(targetDomainIdentifier, targetPath, targetContainer)
      sourceDomain = sourceObserver.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetObserver.SetDomain(targetDomain)
    else:
      absoluteSourcePath = sourceObserver.GetAbsolutePath()
      self.RaiseError("Type mismatch for observer {0:s}\n".format(absoluteSourcePath.ToString()))
      return S2MLFlattenerAction.FAILED
    actualTargetContainer = targetObserver.GetContainer()
    sourceExpression = sourceObserver.GetDefinition()
    targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, targetContainer, actualTargetContainer)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetObserver.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED

# 3.7) Flattening of models
# -------------------------

  def FlattenElement(self, action):
    sourceElement = action.GetDatum(1)
    if sourceElement.GetType()==JanosType.VARIABLE:
      status = self._FlattenVariable(action)
    elif sourceElement.GetType()==JanosType.ASSIGNMENT:
      status = self._FlattenAssignment(action)
    elif sourceElement.GetType()==JanosType.DOUBLE_ASSIGNMENT:
      status = self._FlattenDoubleAssignment(action)
    elif sourceElement.GetType()==JanosType.OBSERVER:
      status = self._FlattenObserver(action)
    else:
      status = S2MLModelFlattener.FlattenElement(self, action)
    return status

  def _FlattenVariable(self, action):
    sourceContainer = action.GetDatum(0)
    sourceVariable = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceVariable.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetVariable = targetContainer.AccessNamedElement(targetPath)
    if targetVariable==None:
      sourceDomain = sourceVariable.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetVariable = self._targetModel.NewVariable(targetDomain, targetPath, targetContainer)
      # targetVariable.SetIdentifierAndInsertInContainerDictionary(targetPath.ToString())
    for name in sourceVariable.GetAttributeNames():
      sourceExpression = sourceVariable.GetAttribute(name)
      targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetVariable.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenAssignment(self, action):
    sourceContainer = action.GetDatum(0)
    sourceAssignment = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourceLeftMember = sourceAssignment.GetLeftMember()
    targetLeftMember = self._FlattenStochasticExpression(sourceContainer, sourceLeftMember)
    if targetLeftMember==None:
      return S2MLFlattenerAction.FAILED
    sourceRightMember = sourceAssignment.GetRightMember()
    targetRightMember = self._FlattenStochasticExpression(sourceContainer, sourceRightMember)
    if targetRightMember==None:
      return S2MLFlattenerAction.FAILED
    targetAssignment = self._targetModel.NewAssignment(targetLeftMember, targetRightMember, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenDoubleAssignment(self, action):
    sourceContainer = action.GetDatum(0)
    sourceAssignment = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourceLeftMember = sourceAssignment.GetLeftMember()
    targetLeftMember = self._FlattenStochasticExpression(sourceContainer, sourceLeftMember)
    if targetLeftMember==None:
      return S2MLFlattenerAction.FAILED
    sourceRightMember = sourceAssignment.GetRightMember()
    targetRightMember = self._FlattenStochasticExpression(sourceContainer, sourceRightMember)
    if targetRightMember==None:
      return S2MLFlattenerAction.FAILED
    targetAssignment = self._targetModel.NewDoubleAssignment(targetLeftMember, targetRightMember, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenObserver(self, action):
    sourceContainer = action.GetDatum(0)
    sourceObserver = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourcePath = sourceObserver.GetPath()
    actualSourceObserver = sourceContainer.AccessNamedElement(sourcePath)
    if actualSourceObserver==None:
      self.RaiseError("This should not happen")
      return assisFlattenerAction.FAILED
    targetPath = actualSourceObserver.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetIdentifier = targetPath.ToString()
    targetObserver = targetContainer.LookForIdentifiedElement(targetIdentifier)
    if targetObserver==None:
      sourceDomain = sourceObserver.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetObserver = self._targetModel.NewObserver(targetDomain, targetPath, targetContainer)
      targetObserver.SetIdentifierAndInsertInContainerDictionary(targetIdentifier)
    sourceExpression = sourceObserver.GetDefinition()
    targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetObserver.SetDefinition(targetExpression)
    return S2MLFlattenerAction.COMPLETED
