# Janos version 1.2.1
# JanosModelRewriter.py
# by Antoine Rauzy
# Copyrights (c) 2018-2022 AltaRica Association
#
# This file is part of Janos, a pedagogical stochastic simulator.
# Janos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Janos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Janos.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Post-processing
# 4) Definition of references
# 3) Reordering of declarations

# 1) Imported modules
# -------------------

from S2MLCore import *

from JanosModels import *

# 2) Creator
# ----------

class JanosModelRewriter(S2MLModelRewriter):
  UNDEFINED = -1

  def __init__(self, engine):
    S2MLModelRewriter.__init__(self, engine)
    self._parameterIndex = 0
    self._variableIndex = 0
    self._equationIndex = 0
    self._observerIndex = 0
    self._currentBlock = None
    self._errors = 0

# 3) Post-processing
# ------------------

  def PostProcessModel(self, model):
    self._errors = 0
    for element in model.GetElements():
      if element.GetType()==JanosType.BLOCK:
        self.PostProcessBlock(element)
        if self._errors!=0:
          return self._errors
    return self._errors

  def PostProcessBlock(self, block):
    self.DefineReferencesInBlock(block)
    if self._errors!=0:
      return
    self.ReorderElementsInBlock(block)

# 4) Definition of references
# ---------------------------

  def DefineReferencesInBlock(self, block):
    previousBlock = self._currentBlock
    self._currentBlock = block
    for element in block.GetElements():
      if element.GetType()==JanosType.BLOCK:
        self.DefineReferencesInBlock(element)
      elif element.GetType()==JanosType.PARAMETER:
        self.DefineReferencesInParameterDeclaration(element)
      elif element.GetType()==JanosType.VARIABLE:
        self.DefineReferencesInVariableDeclaration(element)
      elif element.GetType()==JanosType.ASSIGNMENT:
        self.DefineReferencesInEquation(element)
      elif element.GetType()==JanosType.DOUBLE_ASSIGNMENT:
        self.DefineReferencesInEquation(element)
      elif element.GetType()==JanosType.OBSERVER:
        self.DefineReferencesInObserverDeclaration(element)
    self._currentBlock = previousBlock

  def DefineReferencesInParameterDeclaration(self, parameter):
    definition = parameter.GetDefinition()
    self.DefineReferencesInStochasticExpression(definition)

  def DefineReferencesInVariableDeclaration(self, variable):
    for name in variable.GetAttributeNames():
      expression = variable.GetAttribute(name)
      self.DefineReferencesInStochasticExpression(expression)

  def DefineReferencesInEquation(self, equation):
    for member in equation.GetMembers():
      self.DefineReferencesInStochasticExpression(member)

  def DefineReferencesInStochasticExpression(self, expression):
    if expression.GetType()==JanosType.REFERENCE_EXPRESSION:
      path = expression.GetReferencePath()
      path.MakeIdentifier()
      element = self._currentBlock.AccessNamedElement(path)
      if element==None:
        self.RaiseError("In block {0:s}, undefined reference {1:s}".format(self._currentBlock.GetIdentifier(), path.ToString()))
        self._errors += 1
        return
      expression.SetReferredElement(element)
      return
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return
    for child in expression.GetChildren():
      self.DefineReferencesInStochasticExpression(child)

  def DefineReferencesInObserverDeclaration(self, observer):
    definition = observer.GetDefinition()
    self.DefineReferencesInStochasticExpression(definition)

# 3) Reordering of declarations
# -----------------------------

  def ReorderElementsInBlock(self, block):
    self._parameterIndex = 0
    self._variableIndex = 0
    self._equationIndex = 0
    self._observerIndex = 0
    self.DefineIndicesInBlock(block)
    if self._errors!=0:
      return
    block.SortElements(JanosModelRewriter._GetKeyOfElement)

  @staticmethod
  def _GetKeyOfElement(element):
    if element.GetType()==JanosType.VARIABLE:
      return element.GetIndex()
    if element.GetType()==JanosType.PARAMETER:
      return element.GetIndex()+10000
    if element.GetType()==JanosType.ASSIGNMENT:
      return element.GetIndex()+20000
    if element.GetType()==JanosType.DOUBLE_ASSIGNMENT:
      return element.GetIndex()+20000
    if element.GetType()==JanosType.OBSERVER:
      return element.GetIndex()+30000
    return 0

  def DefineIndicesInBlock(self, block):
    for element in block.GetElements():
      if element.GetType()==JanosType.VARIABLE:
        element.SetIndex(JanosModelRewriter.UNDEFINED)
      elif element.GetType()==JanosType.PARAMETER:
        element.SetIndex(JanosModelRewriter.UNDEFINED)
      elif element.GetType()==JanosType.ASSIGNMENT:
        element.SetIndex(JanosModelRewriter.UNDEFINED)
      elif element.GetType()==JanosType.DOUBLE_ASSIGNMENT:
        element.SetIndex(JanosModelRewriter.UNDEFINED)
      elif element.GetType()==JanosType.OBSERVER:
        element.SetIndex(JanosModelRewriter.UNDEFINED)
    done = False
    while not done:
      done = True
      for element in block.GetElements():
        if element.GetType()==JanosType.PARAMETER:
          if element.GetIndex()==JanosModelRewriter.UNDEFINED:
            unchanged = self.DefineIndicesInParameter(element)
            done = done and unchanged
        elif element.GetType()==JanosType.ASSIGNMENT:
          if element.GetIndex()==JanosModelRewriter.UNDEFINED:
            unchanged = self.DefineIndicesInAssignment(element)
            done = done and unchanged
        elif element.GetType()==JanosType.DOUBLE_ASSIGNMENT:
          if element.GetIndex()==JanosModelRewriter.UNDEFINED:
            unchanged = self.DefineIndicesInDoubleAssignment(element)
            done = done and unchanged
        elif element.GetType()==JanosType.OBSERVER:
          if element.GetIndex()==JanosModelRewriter.UNDEFINED:
            unchanged = self.DefineIndicesInObserver(element)
            done = done and unchanged
    for element in block.GetElements():
      if element.GetType()==JanosType.VARIABLE and element.GetIndex()==JanosModelRewriter.UNDEFINED:
        self.RaiseError("In block {0:s}, absent or circular definition of variable {1:s}".format(block.GetIdentifier(), element.GetIdentifier()))
        self._errors += 1
    return self._errors

  def DefineIndicesInParameter(self, parameter):
    definition = parameter.GetDefinition()
    if self.AreAllIndicesOfReferencesInStochasticExpressionDefined(definition, True):
      self._parameterIndex += 1
      parameter.SetIndex(self._parameterIndex)
      return False
    return True

  def DefineIndicesInAssignment(self, assignment):
    leftMember = assignment.GetLeftMember()
    leftVariable = leftMember.GetReferredElement()
    if leftVariable.GetIndex()!=JanosModelRewriter.UNDEFINED:
      self._equationIndex += 1
      assignment.SetIndex(self._equationIndex)
      return False
    rightMember = assignment.GetRightMember()
    if self.AreAllIndicesOfReferencesInStochasticExpressionDefined(rightMember, False):
      self._variableIndex += 1
      leftVariable.SetIndex(self._variableIndex)
      self._equationIndex += 1
      assignment.SetIndex(self._equationIndex)
      return False
    return True

  def DefineIndicesInDoubleAssignment(self, assignment):
    leftMember = assignment.GetLeftMember()
    rightMember = assignment.GetRightMember()
    leftVariable = leftMember.GetReferredElement()
    rightVariable = rightMember.GetReferredElement()
    if leftVariable.GetIndex()==JanosModelRewriter.UNDEFINED:
      if rightVariable.GetIndex()==JanosModelRewriter.UNDEFINED:
        return True
      self._variableIndex += 1
      leftVariable.SetIndex(self._variableIndex)
    elif rightVariable.GetIndex()==JanosModelRewriter.UNDEFINED:
      self._variableIndex += 1
      rightVariable.SetIndex(self._variableIndex)
      assignment.RemoveAllMembers()
      assignment.AppendMember(rightMember)
      assignment.AppendMember(leftMember)
    self._equationIndex += 1
    assignment.SetIndex(self._equationIndex)
    assignment.SetType(JanosType.ASSIGNMENT)
    return False

  def DefineIndicesInConnection(self, connection):
    undefinedVariable = None
    for member in connection.GetMembers():
      variable = member.GetReferredElement()
      if variable.GetIndex()==JanosModelRewriter.UNDEFINED:
        if undefinedVariable!=None:
          return True
        undefinedVariable = variable
    self._equationIndex += 1
    assignment.SetIndex(self._equationIndex)
    if undefinedVariable==None:
      return False
    self._variableIndex += 1
    undefinedVariable.SetIndex(self._variableIndex)
    return False

  def DefineIndicesInObserver(self, observer):
    definition = observer.GetDefinition()
    if self.AreAllIndicesOfReferencesInStochasticExpressionDefined(definition, True):
      self._observerIndex += 1
      observer.SetIndex(self._observerIndex)
      return False
    return True

  def AreAllIndicesOfReferencesInStochasticExpressionDefined(self, expression, includeParameters):
    if expression.GetType()==JanosType.REFERENCE_EXPRESSION:
      element = expression.GetReferredElement()
      if element.GetType()==JanosType.VARIABLE:
        return element.GetIndex()!=JanosModelRewriter.UNDEFINED
      if element.GetType()==JanosType.PARAMETER and includeParameters:
        return element.GetIndex()!=JanosModelRewriter.UNDEFINED
      return True
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return True
    for child in expression.GetChildren():
      allDefined = self.AreAllIndicesOfReferencesInStochasticExpressionDefined(child, includeParameters)
      if not allDefined:
        return False
    return True
